#include "generic.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <math.h>

#include "vektorr3.h"

using namespace std;

VektorR3 span_x, span_y;
VektorR3 center;

VektorR3 observer;
VektorR3 target;

VektorR3 sight;
VektorR3 light;
VektorR3 mid_sphere;

inline double sqr(double v)
{
  return v*v;
} // srq

inline void SetPixelColor(SDL_Surface *screen, int x, int y, unsigned char red, unsigned char green, unsigned char blue)
{
  ((unsigned char*)(screen->pixels))[x*3+0 + y*screen->pitch] = blue;
  ((unsigned char*)(screen->pixels))[x*3+1 + y*screen->pitch] = green;
  ((unsigned char*)(screen->pixels))[x*3+2 + y*screen->pitch] = red;
} // SetPixelColor

inline void SetPixel(SDL_Surface* screen, int x, int y)
{
  SetPixelColor(screen, x, y, 255, 255, 255);
} // SetPixel

bool CalcProPlane(VektorR3 B, VektorR3 Z, int Res_x, int Res_y, int angle,  VektorR3 &A, VektorR3 &x, VektorR3 &y)
{
	VektorR3 b;
	float    laenge;

	// Blickvektor berechnen

	b = Z - B;

	if((b[0] != 0.0) || (b[1] != 0.0) || (b[2] != 0.0))
	{
		// Aufhnger A der Ebene berechnen

		laenge = b.Norm2();

		A = B + b/laenge;

		// Berechnung des Spannvektors x

		x[1] = 0.0; // Kein geneigter Betrachter -> Projektionsebene parallel zur X-Z Ebene

		if(b[2] != 0.0)
		{
			x[0] = 1/((float)sqrt((double)(1.0+sqr(b[0]/b[2]))));
			x[2] = (-x[0]*b[0])/b[2];
		} // if
		else
		{
			if(b[0] != 0.0)
			{
				x[0] = 0.0;
				x[2] = 1.0;
			} // if
			else
			{
				x[0] = 1.0;
				x[2] = 0.0;
			} // else
		} // else

		// Berechnung des Spannvektors y

		if(x[2] != 0.0)
		{
			if(b[1] != 0.0)
			{
				y[0] = 1.0/((float)sqrt((double)(1.0+sqr((x[0]*b[2])/(x[2]*b[1])-b[0]/b[1])+sqr(x[0]/x[2]))));
				y[1] = y[0]*((x[0]*b[2])/(x[2]*b[1])-b[0]/b[1]);
				y[2] = (-x[0]*y[0])/x[2];
			} // if
			else
			{
			  y = VektorR3(0.0, 1.0, (-x[0]*y[0])/x[2]);
			} // else
		} // if
		else
		{
			if(b[2] != 0.0)
			{
				y[1] = 1.0/((float)sqrt((double)(1.0+sqr(b[1]/b[2]))));
				y[2] = (-y[1]*b[1])/b[2];
				y[0] = 0.0;
			} // if
			else
			{
				y = VektorR3(0,1,0);
			} // else
		} // else

		// Vektorrichtung bestimmen

		if(y[1] < 0.0)
		{
			y *= -1.0;
		} // if

		if(b[2] <= 0.0)
		{
			if(x[0] < 0.0)
			{
				x *= -1.0;
			} // if
		} // if
		else
		{
			if(x[0] > 0.0)
			{
				x *= -1.0;
			} // if
		} // else

		// Berechnung der Vektorlnge

		laenge = ((float)tan((double)((float)angle/114.591559))*2.0)/(float)Res_x;

		x *= laenge;
		y *= laenge;

		return(true);
	} // if
	else
	  return(false); // Betrachter = Ziel (darf nicht sein)
} // CalcProPlane

VektorR3 CalcReflexion(const VektorR3& in, const VektorR3& normal)
{
  VektorR3 reflexion;

  // reflexion = ...;

  return reflexion;
} // CalcReflexion

// return the parameter of the line through center of projection and the
// offset (midpoint of projection plane). 1 = intersection right at projection
// plane, -1 indicates no intersection, values > 1 indicate intersections
// beyond the intersection plane

double IntersectSphere(const VektorR3& center, double radius, VektorR3& offset, VektorR3& direction)
{
  double s;
  double v;
  
  VektorR3 S = center - offset;

  v = 2*direction*S*2*direction*S - 4*direction*direction*(S*S-radius*radius);

  if(v < 0.0) return -1;

  s = (2*direction*S - sqrt(v))/(2*direction*direction);

  return s;
} // IntersectSphere

void DetermineColor(VektorR3 &observer, VektorR3 &sight, unsigned char &red, unsigned char &green, unsigned char &blue)
{
  VektorR3 reflexion, intersection, incident_light, surf_normal, gaze;
  double   s, specular, diffuse, illumination, shade;

  VektorR3 light(-20, -120, -90);

  s = IntersectSphere(mid_sphere, 60.0, observer, sight);


  if(s > 0.0) {

    VektorR3 cut = observer + s*sight;

    VektorR3 n = cut - mid_sphere;
    n *= 1.0/n.Norm2();

    VektorR3 l = light - cut;
    l *= 1.0/l.Norm2();

    // determine diffuse color
   
    shade = l*n;

    // determine specular reflexion

    VektorR3 r = (2.0*n*l)*n - l;
    r *= 1.0/r.Norm2();

    sight *= 1.0/sight.Norm2();

    double spec = -1.0*r*sight;

    if(spec > 0.0) shade += pow(spec,128.0)*3;

    if(shade < 0.0) shade = 0.0;

    shade += 0.09; // some ambient light exists everywhere

    if(shade > 1.0) shade = 1.0;

    red = shade*255;
    green = shade*255;
    blue = shade*255;
    
  } // if
  else {
    sight *= 1.0/sight.Norm2();
    shade = (0.0+pow((sight*VektorR3(0,-1.0,0)), 4))*5;
    if(shade > 1.0) shade = 1.0;

    red = 0;
    green = 0;
    double fblue = (double)(shade*255.0+48);
    
    if(fblue > 255) fblue = 255;

    blue = fblue;
  } // else

} // DetermineColor    

int main(int argc, char** argv)
{
    SDL_Surface      *screen, *backup_screen;
    SDL_Event        event;
    SDL_Rect         dst_rect, full_image_rect, scrolled_rect, location;
    int              window_width, window_height, thumb_height, thumb_border_size;
    long             pixel_set = 0;

    window_width = 720; window_height = 720;

    if(SDL_Init(SDL_INIT_VIDEO) < 0) {
	cerr << "Could not initialize SDL video" << endl;
	exit(-1);
    } // if
    atexit(SDL_Quit);

    Uint32 flags = 0;

    // flags |= SDL_FULLSCREEN;

    screen = SDL_SetVideoMode(window_width, window_height, 24, flags);

    if(!screen) {
	cerr << "could not SetVideoMode() - abort" << endl;
	exit(-1);
    } // if

    unsigned char red, green, blue;

    observer = VektorR3(90, 40, -90);
    target = VektorR3(0, 0, 90);

    mid_sphere = VektorR3(0, 0, 90);

    CalcProPlane(observer, target, window_width, window_height, 
		 60, center, span_x, span_y);
   
    for(int y = 0; y < window_height; y++) {
      for(int x = 0; x < window_width; x++) {

	sight = (center + ((double)(x-window_width/2)*span_x + (double)(y-window_height/2)*span_y)) - observer;

	DetermineColor(observer, sight, red, green, blue);

	SetPixelColor(screen, x, y, red, green, blue);

      } // for

      SDL_UnlockSurface(screen);
      SDL_UpdateRect(screen, 0, y, window_width, 1);
      SDL_LockSurface(screen);
    } // for

    for(;;) {

	if((SDL_PollEvent(&event) == 0) && (event.type == SDL_KEYDOWN)) 
	    goto exit_app;

	switch(event.type) {
		
	    case SDL_MOUSEMOTION:	      
		break;
		
	    case SDL_MOUSEBUTTONDOWN:
		break;
		
	    case SDL_MOUSEBUTTONUP:
		break;
		
	    case SDL_KEYDOWN:
		
		switch(event.key.keysym.sym) {
		    
		    case 27:
		    case 'q':
		    case 'Q':
			goto exit_app;
			
		    default:
			break;
		} // switch
	} // switch

    } // for
	    
 exit_app:

    SDL_UnlockSurface(screen);	    
    SDL_FreeSurface(screen);

    exit(0);
} // main
